"""Rule to test the --custom-platform-rules-dir Click option."""

from mc_nag.base_utils.models.rule import (
    BaseRule,
    WARNING as RULE_WARNING
)


class CustomPlatformRule(BaseRule):
    """Create a custom rule to test --custom-platform-rules-dir option."""

    rule_id = 'custom-platform-rule-test'
    description = """Testing the --custom-platform-rules-dir option."""
    severity = RULE_WARNING
    url = 'https://github.com/stelligent/mc-nag/tests/rules/custom/' + \
          'custom_rule_for_option_testing.py'
    resolution = 'N/A'
    category_tags = {"custom-resource"}
    source_tags = {"stelligent"}

    def evaluate(self):
        """Find storageAccount resources and validate 'encryption' property.

        1. Obtain a list of storageAccount resources.
        2. Search their properties for 'encryption'.
        3. Ensure at least one service is enabled.
        4. Ensure keySource subproperty is set.
        """
        return self.template_model.resources



"""Exceptions thrown for template processing."""

import sys


class QuietException(Exception):
    """Output only the exception name and message, no traceback."""


class InvalidTemplate(QuietException):
    """Raise this exception when the template is malformed."""


class InvalidParameter(QuietException):
    """Raise this exception when a parameter is missing a value."""


class VariableKeyError(QuietException):
    """Raise this exception when a variable key does not have a value."""


# No unit tests for this, due to Python mechanics.
# https://stackoverflow.com/a/46351418
def quiet_exception_hook(kind, message, traceback):
    """Print only exception name and message if based on QuietException."""
    if QuietException in kind.__bases__:
        # Only print Error Type and Message
        print(f'{kind.__name__}: {message}')
    else:
        # Print Error Type, Message and Traceback
        sys.__excepthook__(kind, message, traceback)



"""Metaclasses to help with base class creation."""


class RequiredAttributeDefinitionMeta(type):
    """Metadata class to enforce required attributes on subclasses."""

    def __call__(cls, *args, **kwargs):
        """Execute check_required_attributes() when class is called."""
        class_object = type.__call__(cls, *args, **kwargs)
        class_object.check_required_attributes()
        return class_object



"""Tests for rule evaluation logic."""

import re
from pytest import fixture, raises
from mc_nag.base_utils import read_template_file
from mc_nag.base_utils.models.rule import (
    BaseRule,
    ERROR as RULE_ERROR,
    WARNING as RULE_WARNING
)
from mc_nag.base_utils.rule_evaluator import RuleEvaluator
from tests.rules.standard.standard_rule import StandardRule
from tests.rules.custom.custom_rule_for_option_testing import \
    CustomPlatformRule
from tests.fixtures import TestTemplate

# pylint: disable=redefined-outer-name

TEMPLATES_PATH = 'tests/templates'
VALID_TEMPLATE = f'{TEMPLATES_PATH}/parser_test_good.json'
INVALID_TEMPLATE = f'{TEMPLATES_PATH}/parser_test_bad.json'


@fixture(scope='module')
def valid_template_model():
    """Return template model for a valid template."""
    return TestTemplate(read_template_file(VALID_TEMPLATE))


@fixture(scope='module')
def invalid_template_model():
    """Return template model for an invalid template."""
    return TestTemplate(read_template_file(INVALID_TEMPLATE))


def test_rule_evaluator_standard_rules():
    """Happy Path: Ensure we can import standard rules."""
    rule_evaluator = RuleEvaluator(None, 'tests/rules/standard')

    assert StandardRule.__name__ in \
        [rule.__name__ for rule in rule_evaluator.rule_set]


def test_rule_evaluator_custom_rules():
    """Happy Path: Ensure we can import custom rules."""
    rule_evaluator = RuleEvaluator(None, 'tests/rules/standard', True,
                                   ['tests/rules/custom'])

    rule_names = [rule.__name__ for rule in rule_evaluator.rule_set]
    assert StandardRule.__name__ in rule_names
    assert CustomPlatformRule.__name__ in rule_names


def test_rule_evaluator_only_custom_rules():
    """Happy Path: Ensure we can import only custom rules."""
    rule_evaluator = RuleEvaluator(None, '', False,
                                   ['tests/rules/custom'])

    rule_names = [rule.__name__ for rule in rule_evaluator.rule_set]
    assert StandardRule.__name__ not in rule_names
    assert CustomPlatformRule.__name__ in rule_names


def test_rule_evaluator_no_rules():
    """Sad Path: Ensure we can handle the case when no rules are supplied.

    This should never happen, as it is handled by Click in main.py,
    but good to validate we will get an empty rule_set.
    """
    rule_evaluator = RuleEvaluator(None, None, False)

    assert not rule_evaluator.rule_set


def test_rule_evaluator_passing_template(valid_template_model):
    """Happy Path: Ensure we can evaluate a rule against a valid template."""
    rule_evaluator = RuleEvaluator(valid_template_model,
                                   'tests/rules/standard')
    rule_evaluator.rule_set = set([StandardRule])
    violations_by_rule, violation_counts, rules = rule_evaluator.evaluate_rules()

    assert not violations_by_rule
    assert not violation_counts
    assert rules == {StandardRule}


def test_rule_evaluator_failing_rule(invalid_template_model):
    """Sad Path: Ensure we can evaluate a rule which returns violations."""
    rule_evaluator = RuleEvaluator(invalid_template_model,
                                   'tests/rules/custom')
    rule_evaluator.rule_set = set([CustomPlatformRule])
    violations_by_rule, violation_counts, rules = rule_evaluator.evaluate_rules()

    assert CustomPlatformRule in violations_by_rule
    assert len(violations_by_rule[CustomPlatformRule]) == 1
    assert RULE_WARNING in violation_counts
    assert violation_counts[RULE_WARNING] == 1
    assert rules == {CustomPlatformRule}


def test_rule_evaluator_bad_rule(invalid_template_model):
    """Sad Path: Ensure we raise an exception for a bad rule class."""
    class BadRule(BaseRule):  # pylint: disable=abstract-method
        """Bad rule for testing."""

        rule_id = 'BAD'
        description = 'Bad rule.'
        severity = RULE_ERROR
        category_tags = {"custom-resource"}
        source_tags = {"stelligent"}

    rule_evaluator = RuleEvaluator(invalid_template_model,
                                   'tests/rules/standard')
    rule_evaluator.rule_set = set([BadRule])

    with raises(NotImplementedError, match="'BadRule' must define"):
        _, _, _ = rule_evaluator.evaluate_rules()


def test_rule_evaluator_display_rules_no_dupes(capsys):
    """Happy Path: Display rules without duplicates."""
    RuleEvaluator(None, 'tests/rules/standard').display_rules()

    captured = capsys.readouterr()
    assert re.search('No duplicates', captured.out)
    assert re.search(r'standard-rule.*StandardRule', captured.out)


def test_rule_evaluator_display_rules_with_dupes(capsys):
    """Happy Path: Display rules with duplicates."""
    RuleEvaluator(None, 'tests/rules/standard', True,
                  ['tests/rules/custom']).display_rules()

    captured = capsys.readouterr()
    assert re.search('Found duplicate rule IDs!', captured.out)
    assert re.search(r'Duplicate Rule ID.*Duplicate Rule Names', captured.out)
    assert re.search(
        r'DuplicateCustomPlatformRule \(duplicate_custom_rule_for_testing.py',
        captured.out
    )



"""Describe a generic rule class."""
import warnings

from mc_nag.base_utils.metaclasses import RequiredAttributeDefinitionMeta

ERROR = 'ERROR'
WARNING = 'WARNING'
STYLE = 'STYLE'


class BaseRule(metaclass=RequiredAttributeDefinitionMeta):
    """Define an extensible base rule class with required attributes.

    One can subclass this BaseRule class as a normal python subclass, but also
    must ensure the required attributes are included and the evaluate()
    function is overridden.

    Example:
    | from base_utils.models.rule import (
    |     BaseRule,
    |     WARNING as RULE_WARNING
    | )
    |
    | def NewRule(BaseRule):
    |     '''This is a new rule.'''
    |
    |     rule_id = '001'
    |     description = 'This is a new rule.'
    |     severity = RULE_WARNING
    |     url = 'https://rule.info/'
    |     resolution = 'Instructions how to fix the violations.'
    |
    |     def evaluate(self):
    |         '''Logic to implement the rule.'''
    |         pass

    To instantiate the rule, the template model must be the only argument.

    Example:
    | rule_object = NewRule(template_model)


    :param template_model: Data model object from the associated platform's
                           template class.

    :attr rule_id: Unique identifier for the rule.
    :attr description: Short description of what the rule is checking.
    :attr severity: Category of the violation (ERROR, WARNING, or STYLE).
    :attr url: Link to more information about the rule.
    :attr resolution: Instructions on how to fix violations of the rule.
    :attr category_tags: Set by the Rule author.
    :attr source_tags: Set by the Rule author.
    :attr clean_category_tags: Validated category names
          (e.g. "storage", "security", "networking")
    :attr clean_source_tags: Validated source names
          (e.g. "company_name", "team_name", "creator_name")
    """

    rule_id = None
    description = None
    severity = None
    url = None
    resolution = None
    category_tags = None
    source_tags = None
    required_attributes = [
        'rule_id',
        'description',
        'severity',
        'url',
        'resolution'
    ]

    def __init__(self, template_model):
        """Assign template_model as an attribute."""
        self.template_model = template_model
        self.__set_tags()

    def check_required_attributes(self):
        """Ensure required properties are set."""
        missing_attributes = [attr for attr in self.required_attributes
                              if getattr(self, attr) is None]
        if missing_attributes:
            raise NotImplementedError(
                f'Class \'{self.__class__.__name__}\' must define ' +  # noqa: W504
                f'{missing_attributes} attribute(s)!'
            )

    def evaluate(self):
        """Implement rule logic.

        Any logic needed to detect violating resources should be implemented
        in this function.

        :return: List of violating resource objects.
        """
        raise NotImplementedError(
            f'Class \'{self.__class__.__name__}\' must implement an ' +  # noqa: W504
            'evaluate() method with logic to implement rule!'
        )

    def __clean_tags(self, dirty_tags):  # pylint: disable=no-self-use
        """Sanitize tags for consistency.

        :return: A set data type of lowercase strings.
        """
        result = set(x.lower() for x in dirty_tags if isinstance(x, str))
        return result

    def __process_tags(self, evaluate_tags):
        """Type checks tags.

        Issues a warning if no tags are set.

        :return: Set of lowercase string tags or an empty set.
        """
        if isinstance(evaluate_tags, set):
            result = self.__clean_tags(evaluate_tags)
        else:
            warnings.warn(
                f"""Class \'{self.__class__.__name__}\' tags properties are optional.
                    When configured, use a Set of lowercase
                    strings. e.g. {{"storage"}}""", UserWarning, stacklevel=2)
            result = set()
        return result

    def __set_tags(self):
        """Assign sanitized tags."""
        self.clean_category_tags = self.__process_tags(self.category_tags)
        self.clean_source_tags = self.__process_tags(self.source_tags)



"""Utility functions related to rule evaluation."""

import os
import sys
from importlib.util import module_from_spec, spec_from_file_location
from itertools import chain
from .printers import BCOLORS


class RuleEvaluator:
    """Engine to evaluate rules.

    This class will find all rules inside the supplied standard rules
    directory and evaluate them against the given template model.
    :param model: Template data model built from a template.
    """

    def __init__(self, model, standard_rules_dir, enable_standard_rules=True,
                 custom_rules_dirs=None):
        """Assign arguments as attributes and prepare rules."""
        self.model = model
        self._standard_rules_dir = standard_rules_dir
        self.rule_set = self._retrieve_rule_objects(enable_standard_rules,
                                                    custom_rules_dirs)

    def evaluate_rules(self):
        """Evaluate given rules against the data model."""
        violations_by_rule = {}
        severity_counts = {}

        for rule in self.rule_set:
            # Instantiate rule
            rule_object = rule(self.model)

            # Evaluate rule
            violating_resources = rule_object.evaluate()

            # Continue to next rule if no violations found
            if not violating_resources:
                continue

            # Add to violations by rule
            if rule not in violations_by_rule:
                violations_by_rule[rule] = []
            violations_by_rule[rule] += violating_resources

            # Add to severity counter
            if rule_object.severity not in severity_counts:
                severity_counts[rule_object.severity] = 0
            severity_counts[rule_object.severity] += 1

        return violations_by_rule, severity_counts, self.rule_set

    def display_rules(self):
        """Output list of rules and rule IDs in given format.

        :param output: Format in which the output should be.
        :param verbose: Level of verbosity.
        """
        # Dict to hold values for duplicate checking
        dupe_rules = {}
        rev_dict = {}

        # Gather table formatting information & print header
        left_pad = max(max([len(rule.rule_id) for rule in self.rule_set]), 7)
        right_pad = max(max([len(rule.__name__) for rule in self.rule_set]), 9)
        print(f'{"Rule ID":{left_pad}} | Rule Name')
        print(f"{''.rjust(left_pad, '-')} | {''.rjust(right_pad, '-')}")

        # Iterate over all found rules and build rev dict
        for rule in sorted(self.rule_set, key=lambda x: x.rule_id):
            print(f'{rule.rule_id:{left_pad}} | {rule.__name__}')
            rev_dict.setdefault(rule.rule_id, set()).add(
                f'{rule.__name__} ({rule.__module__}.py)'
            )

        # Find duplicate rule IDs
        for key, values in rev_dict.items():
            if len(values) > 1:
                dupe_rules[key] = values

        # Output
        if dupe_rules:
            left_pad = max(max([len(key) for key in dupe_rules]), 17)
            print(f'\n{BCOLORS["ERROR"]}Found duplicate rule IDs!' +  # noqa: W504
                  f'{BCOLORS["ENDC"]}\n')
            print(f'{"Duplicate Rule ID":{left_pad}} | Duplicate Rule Names')
            for key, values in dupe_rules.items():
                print(f"{''.rjust(left_pad, '-')} | {''.rjust(20, '-')}")
                print(f"{key:{left_pad}} | " +  # noqa: W504
                      f',\n{"".rjust(left_pad)} | '.join(values))
        else:
            print(f'\n{BCOLORS["OKGREEN"]}No duplicates{BCOLORS["ENDC"]} 🎉')

        return dupe_rules

    def _retrieve_rule_objects(self, enable_standard_rules=True,
                               custom_rules_dirs=None):
        """Return list of rule classes to evaluate.

        Based on this StackOverflow info:
        https://stackoverflow.com/a/41904558
        """
        paths = []

        # Platform-specific standard rules path
        if enable_standard_rules:
            paths.append(self._standard_rules_dir)

        # Add custom rules directories to paths if existent
        if custom_rules_dirs:
            paths += custom_rules_dirs

        # Set to hold found rules
        rules = set()
        found_rule_modules = set()

        # Iterate over all rules paths (standard and custom)
        for path, _, files in chain.from_iterable(os.walk(path)
                                                  for path in paths):
            # Iterate over .py files in path
            for py_file in [f[:-3] for f in files
                            if f.endswith('.py') and f != '__init__.py']:
                # Prepare module path to import
                module_path = f'{path}/{py_file}.py'

                # Import .py file as a module
                spec = spec_from_file_location(py_file, module_path)
                mod = module_from_spec(spec)
                spec.loader.exec_module(mod)

                # Retrieve list of classes from module
                classes = [getattr(mod, x) for x in dir(mod)
                           if isinstance(getattr(mod, x), type)
                           and x != 'BaseRule']  # noqa: W503

                # Set new found class names as part of the system modules
                for cls in classes:
                    setattr(sys.modules[__name__], cls.__name__, cls)

                # Add new found classes to rules set
                if py_file not in found_rule_modules:
                    rules.update(classes)
                    found_rule_modules.add(py_file)

        return rules



"""BaseParser class."""

import re
from json.decoder import JSONDecodeError
import jstyleson
from yaml import safe_load
from yaml.scanner import ScannerError as YamlScannerError
from yaml.parser import ParserError as YamlParserError
from mc_nag.base_utils.exceptions import InvalidTemplate


class BaseParser:
    """Parser class to interpret raw templates."""

    def __init__(self, template_string, parameter_file=None):
        """Init parser object with template and parsed template."""
        self.template_string = template_string
        self.parsed_template = self.parse_raw_file(self.template_string)
        self.validate_file(file_type='Template', parsed_contents=self.parsed_template)
        self.file_lines = template_string.split('\n')
        self.parameter_file = parameter_file
        self.parsed_parameter_file = self.parse_raw_file(self.parameter_file)
        self.validate_file(file_type='Parameter', parsed_contents=self.parsed_parameter_file)

    @staticmethod
    def parse_raw_file(file_contents=None):
        """Return parsed dictionary."""
        if file_contents is not None:
            try:
                return safe_load(file_contents)
            except (YamlParserError, YamlScannerError):
                try:
                    return jstyleson.loads(file_contents)
                except JSONDecodeError:
                    raise InvalidTemplate('Invalid YAML or JSON syntax!')
        else:
            return None

    @staticmethod
    def validate_file(file_type=None, parsed_contents=None, required_keys=None,
                      exception_class=InvalidTemplate):
        """Check template string for basic requirements."""
        if required_keys and not all(key in parsed_contents
                                     for key in required_keys):
            missing_keys = ', '.join([key for key in required_keys
                                      if key not in parsed_contents])
            message = f'{file_type} file is missing the {missing_keys} key(s)!'
            raise exception_class(message)

    def find_line_number(self, search_pattern: str) -> int:
        """Return line number from original template from search pattern."""
        regex_pattern = re.compile(re.escape(search_pattern))
        matched_lines = list(filter(regex_pattern.search,
                                    self.file_lines))
        try:
            return self.file_lines.index(matched_lines[0]) + 1
        except IndexError:
            return -1

    @property
    def resources(self):
        """Return parsed resources dict."""
        return self.parsed_template['resources']

    @property
    def parameters(self):
        """Return parsed parameters dict."""
        return self.parsed_template.get('parameters', {})

    @property
    def variables(self):
        """Return parsed variables dict."""
        return self.parsed_template.get('variables', {})

    @property
    def functions(self):
        """Return parsed functions dict."""
        return self.parsed_template.get('functions', {})

    @property
    def outputs(self):
        """Return parsed outputs dict."""
        return self.parsed_template.get('outputs', {})



"""Tests for BaseTemplate."""

from pytest import raises
from mc_nag.base_utils.models.template import BaseTemplate

TEMPLATE_PATH = 'tests/templates/base'


def test_base_template_good_subclass():
    """Happy Path: Ensure subclass can create with required attributes."""
    class GoodSubclassBaseTemplate(BaseTemplate):
        """Create subclass of BaseTemplate with required attributes."""

        def __init__(self, template_string):
            """Init GoodSubclassBaseTemplate."""
            self.template_string = template_string
            self.parsed_template = 'This is a good parsed template'

        @property
        def resources(self):
            """List of template resources."""
            return []

        @property
        def parameters(self):
            """List of template parameters."""
            return []

        @property
        def functions(self):
            """List of template functions."""
            return []

        @property
        def variables(self):
            """List of template variables."""
            return []

        @property
        def outputs(self):
            """List of template outputs."""
            return []

    good_template_string = 'Good template string'
    good_object = GoodSubclassBaseTemplate(good_template_string)

    assert len(good_object.resources) == 0
    assert good_object.template_string == good_template_string
    assert good_object.parsed_template == 'This is a good parsed template'


def test_base_template_bad_subclass():
    """Sad Path: Catch error when subclass is missing attributes."""
    # pylint: disable=too-few-public-methods
    class BadSubclassBaseTemplate(BaseTemplate):
        """Create subclass of BaseTemplate without setting all attributes."""

        def __init__(self, template_string):
            """Init BadSubclassBaseTemplate."""
            self.wrong_template_string = template_string
            self.wrong_parsed_template = 'This is a bad parsed template'

    bad_template_string = 'Bad subclass'

    with raises(NotImplementedError,
                match=r"must define.*attribute\(s\) as type 'list'"):
        _ = BadSubclassBaseTemplate(bad_template_string)



"""Tests for output printers."""

import re
from pytest import fixture, raises
from mc_nag.base_utils import read_template_file
from mc_nag.base_utils.models.rule import WARNING as RULE_WARNING
from mc_nag.base_utils.printers import OutputPrinter
from tests.rules.custom.custom_rule_for_option_testing import \
    CustomPlatformRule
from tests.fixtures import TestResource, TestTemplate

TEMPLATES_PATH = 'tests/templates'
VALID_TEMPLATE = f'{TEMPLATES_PATH}/parser_test_good.json'

# pylint: disable=redefined-outer-name


@fixture(scope="module")
def valid_template_evaluation():
    """Return violations after evaluating a valid template."""
    return ({}, {}, [CustomPlatformRule])


@fixture(scope="module")
def invalid_template_evaluation():
    """Return violations after evaluating an invalid template."""
    return (
        {CustomPlatformRule: [
            TestResource(
                resource_type='Microsoft.Storage/storageAccounts',
                resource_name='insecurestorageaccount',
                properties={'supportsHttpsTrafficOnly': True},
                raw_string="{'type': 'Microsoft.Storage/storageAccounts', " +
                "'apiVersion': '2019-04-01', 'name': " +
                "'insecurestorageaccount', 'location': 'eastus', 'sku':" +
                " {'name': 'Standard_LRS'}, 'kind': 'StorageV2', " +
                "'properties': {'supportsHttpsTrafficOnly': True}}",
                line_number=8
            )
        ]},
        {'WARNING': 1},
        [CustomPlatformRule]
    )


def test_template_output_good_template(capsys):
    """Happy Path: Ensure we can print a proper template."""
    template_model = TestTemplate(read_template_file(VALID_TEMPLATE))
    print(template_model)

    captured = capsys.readouterr()
    assert re.search('RESOURCES', captured.out)
    assert re.search('Resource Line Number: 45', captured.out)


def test_template_output_bad_template(capsys):
    """Sad Path: Ensure AttributeError is raised with bad template."""
    template_model = None
    print(template_model)

    captured = capsys.readouterr()
    assert re.match(r'^None$', captured.out)


def test_text_printer_no_violations(capsys, valid_template_evaluation):
    """Happy Path: Ensure proper output is printed with no violations."""
    violations_dict, violations_count, rules = valid_template_evaluation
    OutputPrinter.text(violations_dict, violations_count, rules)

    captured = capsys.readouterr()
    assert re.search('No violations', captured.out)


def test_text_printer_with_violations(capsys, invalid_template_evaluation):
    """Happy Path: Ensure proper output is printed with violations."""
    violations_dict, violations_count, rules = invalid_template_evaluation
    OutputPrinter.text(violations_dict, violations_count, rules)

    captured = capsys.readouterr()
    assert re.search('CustomPlatformRule', captured.out)
    assert re.search(fr'{RULE_WARNING}.* : 1', captured.out)


def test_text_printer_bad_violations_data():
    """Happy Path: Ensure proper output is printed with bad."""
    with raises(AttributeError, match="'str' object has no attribute"):
        OutputPrinter.text({'bad': 'data'}, {'bad': 'data'}, [])


def test_json_printer_no_violations(capsys, valid_template_evaluation):
    """Happy Path: Ensure proper output is printed with no violations."""
    violations_dict, violations_count, rules = valid_template_evaluation
    OutputPrinter.json(violations_dict, violations_count, rules)

    captured = capsys.readouterr()
    assert re.search(r'"violations": \[\]', captured.out)
    assert re.search(r'"report": {.*"total_violations": 0', captured.out)
    assert re.search(r'"report": {.*"rules_evaluated": \["CustomPlatformRule"\]',
                     captured.out)


def test_json_printer_with_violations(capsys, invalid_template_evaluation):
    """Happy Path: Ensure proper output is printed with violations."""
    violations_dict, violations_count, rules = invalid_template_evaluation
    OutputPrinter.json(violations_dict, violations_count, rules)

    captured = capsys.readouterr()
    assert re.search(r'"violations": \[{"name": "CustomPlatformRule',
                     captured.out)
    assert re.search(r'"report": {.*"total_violations": 1, "WARNING": 1',
                     captured.out)


def test_json_printer_bad_violations_data():
    """Happy Path: Ensure proper output is printed with bad data."""
    with raises(AttributeError, match="'str' object has no attribute"):
        OutputPrinter.json({'bad': 'data'}, {'bad': 'data'}, [])


def test_yaml_printer_no_violations(capsys, valid_template_evaluation):
    """Happy Path: Ensure proper output is printed with no violations."""
    violations_dict, violations_count, rules = valid_template_evaluation
    OutputPrinter.yaml(violations_dict, violations_count, rules)

    captured = capsys.readouterr()
    assert re.search(r'violations:', captured.out)
    assert re.search(r'^report:\n(.*\n)*\s+total_violations: 0',
                     captured.out, re.MULTILINE)
    assert re.search(r'^report:\n(.*\n)*\s+rules_evaluated:\n\s+- CustomPlatformRule',
                     captured.out, re.MULTILINE)


def test_yaml_printer_with_violations(capsys, invalid_template_evaluation):
    """Happy Path: Ensure proper output is printed with violations."""
    violations_dict, violations_count, rules = invalid_template_evaluation
    OutputPrinter.yaml(violations_dict, violations_count, rules)

    captured = capsys.readouterr()
    assert re.search(r"  name: CustomPlatformRule", captured.out)
    assert re.search(r"violations:\n- description: Testing the .*\n",
                     captured.out, re.MULTILINE)
    assert re.search(r'report:\n(.*\n)*\s+WARNING: 1',
                     captured.out, re.MULTILINE)
    assert re.search(r'report:\n(.*\n)*\s+total_violations: 1',
                     captured.out, re.MULTILINE)
    assert re.search(r'report:\n(.*\n)*\s+rules_evaluated:\n\s+- CustomPlatformRule',
                     captured.out, re.MULTILINE)


def test_yaml_printer_bad_violations_data():
    """Happy Path: Ensure proper output is printed with bad data."""
    with raises(AttributeError, match="'str' object has no attribute"):
        OutputPrinter.yaml({'bad': 'data'}, {'bad': 'data'}, [])


def test_none_printer_no_violations(capsys, valid_template_evaluation):
    """Happy Path: Ensure proper output is printed with no violations."""
    violations_dict, violations_count, rules = valid_template_evaluation
    OutputPrinter.none(violations_dict, violations_count, rules)

    captured = capsys.readouterr()
    assert not captured.out


def test_none_printer_with_violations(capsys, invalid_template_evaluation):
    """Happy Path: Ensure proper output is printed with violations."""
    violations_dict, violations_count, rules = invalid_template_evaluation
    OutputPrinter.none(violations_dict, violations_count, rules)

    captured = capsys.readouterr()
    assert not captured.out


def test_none_printer_bad_violations_data(capsys):
    """Happy Path: Ensure proper output is printed with bad data."""
    OutputPrinter.none({'bad': 'data'}, {'bad': 'data'}, [])

    captured = capsys.readouterr()
    assert not captured.out



"""Standard rule for testing."""

from mc_nag.base_utils.models.rule import (
    BaseRule,
    WARNING as RULE_WARNING
)


class StandardRule(BaseRule):
    """Create a standard rule for testing."""

    rule_id = 'standard-rule'
    description = """Testing standard rules."""
    severity = RULE_WARNING
    url = 'https://github.com/stelligent/mc-nag/tests/rules/standard/' + \
          'standard_rule.py'
    resolution = 'N/A'
    category_tags = {"custom-resource"}
    source_tags = {"stelligent"}

    def evaluate(self):
        """Return empty list."""
        return []



"""Describe a generic template in Python memory."""

from mc_nag.base_utils.metaclasses import RequiredAttributeDefinitionMeta


# pylint: disable=too-few-public-methods
class BaseTemplate(metaclass=RequiredAttributeDefinitionMeta):
    """Define an extensible base template class with required attributes.

    One can subclass this BaseTemplate class as a normal python subclass, but
    also must ensure the required attributes are included.

    Example:
    | from base_utils.models.template import BaseTemplate
    |
    | def NewTemplate(BaseTemplate):
    |     '''This is a new template.'''
    |
    |     def __init__(self, new_template_string):
    |         '''Initialze NewTemplate object.'''
    |         self.template_string = new_template_string
    |         self.parsed_template = NewParser(new_template_string)
    |
    |     @property
    |     def resources(self):
    |         '''Return list of resources in the template.'''
    |         return [resource for resources in ...]
    |
    |     ...

    :attr template_string: Raw string read from the template file.
    :attr parsed_template: Output from the platform's parser object.
    :attr resources: List of platform-specific resources in the template.
    :attr parameters: List of platform-specific parameters in the template.
    :attr outputs: List of platform-specific outputs in the template.
    :attr functions: List of platform-specific functions in the template.
    :attr variables: List of platform-specific variables in the template.
    """

    template_string = None
    parsed_template = None
    resources = None
    parameters = None
    outputs = None
    functions = None
    variables = None
    required_attributes = [
        'template_string',
        'parsed_template',
        'resources',
        'parameters',
        'outputs',
        'functions',
        'variables'
    ]

    def check_required_attributes(self):
        """Ensure required properties are set."""
        missing_attributes = [
            attr for attr in self.required_attributes
            if attr not in ['template_string', 'parsed_template'] and  # noqa: W504
            not isinstance(getattr(self, attr), list)
        ]
        if not self.template_string:
            missing_attributes.append('template_string')
        if not self.parsed_template:
            missing_attributes.append('parsed_template')

        if missing_attributes:
            raise NotImplementedError(
                f'Class \'{self.__class__.__name__}\' must define ' +  # noqa: W504
                f'{missing_attributes} attribute(s) as type \'list\''
            )

    def __repr__(self):
        """Print debugging output for template models.

        :param template_model: Data model of a platform's template.
        """
        item_divider_char = '+'
        item_divider_length = 15
        item_divider = ''.rjust(item_divider_length, item_divider_char)

        # Resources
        repr_string = _repr_section_divider('RESOURCES')
        for res in self.resources:
            repr_string += f"""Resource Name: {res.resource_name}
Resource Type: {res.resource_type}
Resource Properties: {res.properties}
Resource Line Number: {res.line_number}
{item_divider}\n"""

        # Parameters
        repr_string += _repr_section_divider('PARAMETERS')
        for param in self.parameters:
            repr_string += f"""Parameter Name: {param.parameter_name}
Parameter Value: {param.parameter_value}
Parameter Type: {param.parameter_type}
{item_divider}\n"""

        # Outputs
        repr_string += _repr_section_divider('OUTPUTS')
        for output in self.outputs:
            repr_string += f"""Output Name: {output.output_name}
Output Value: {output.output_value}
Output Type: {output.output_type}
{item_divider}\n"""

        # Functions
        repr_string += _repr_section_divider('FUNCTIONS')
        for function in self.functions:
            repr_string += f"""Function Namespace: {function.function_namespace}
Function Members: {function.function_members}
{item_divider}\n"""

        # Variables
        repr_string += _repr_section_divider('VARIABLES')
        for variable in self.variables:
            repr_string += f"""Variable Name: {variable.variable_name}
Variable Value: {variable.variable_value}
{item_divider}\n"""
        repr_string += _repr_section_divider()

        return repr_string


def _repr_section_divider(title=None):
    """Return formatted string for __repr__ section dividers."""
    divider_char = '='
    divider_length = 53
    divider = ''.rjust(divider_length, divider_char)

    if title:
        return f"{divider}\n{title}\n{divider}\n"

    return divider



"""Click CLI Module."""
import click

# pylint: disable=too-many-arguments

COMMON_OPTIONS = [
    click.option('--enable-standard-rules/--disable-standard-rules',
                 default=True,
                 help='Allows for enabling/disabling the standard rule set ' +  # noqa: W504
                 'that ships with mc-nag.'),
    click.option('--custom-platform-rules-dir', '-C', multiple=True,
                 type=click.Path(exists=True), default=None,
                 help='Path to a directory containing custom rules. ' +  # noqa: W504
                 'Allows multiple.')
]
MAIN_OPTIONS = COMMON_OPTIONS + [
    click.option('--rules', is_flag=True),
    click.option('--filepath', '-f', type=click.Path(exists=True),
                 required=False),
    click.option('--output', '-o', default='text',
                 type=click.Choice(['text', 'json', 'yaml', 'none'],
                                   case_sensitive=False)),
    click.option('--paramfile', '-p', type=click.Path(exists=True)),
    click.option('--verbose', '-v', count=True)
]


def add_click_options(options):
    """Decorate with Click options."""

    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options



"""Tests for BaseRule."""

from pytest import raises, warns
from mc_nag.base_utils.models.rule import BaseRule

TEMPLATE_MODEL = {'raw': 'model'}
OBJECT_ID = '001'


def test_base_rule_good_subclass():
    """Happy Path: Ensure subclass can create with required attributes."""
    class GoodSubclassBaseRule(BaseRule):
        """Create subclass of BaseRule with required attributes."""

        rule_id = OBJECT_ID
        description = 'This is a rule that properly instantiates BaseRule.'
        severity = 'FAIULRE'
        url = 'http://rule.documentation/'
        resolution = 'Take these actions to fix the violation.'
        category_tags = {"storage"}
        source_tags = {"stelligent"}

        def evaluate(self):
            """Logic to implement the rule."""
            return self.template_model

    good_object = GoodSubclassBaseRule(TEMPLATE_MODEL)

    assert good_object.rule_id == OBJECT_ID
    assert good_object.evaluate() == TEMPLATE_MODEL


def test_base_rule_bad_subclass():
    """Sad Path: Catch error when subclass is missing attributes."""
    # pylint: disable=too-few-public-methods
    class RuleMissingAttributes(BaseRule):
        """Create subclass of BaseRule without setting all attributes."""

        description = 'This is a rule that properly instantiates BaseRule.'
        severity = 'FAIULRE'
        url = 'http://rule.documentation/'
        category_tags = {"storage"}
        source_tags = {"stelligent"}

    with raises(NotImplementedError,
                match=r"must define.*attribute\(s\)!"):
        _ = RuleMissingAttributes(TEMPLATE_MODEL)

    class RuleMissingEvaluate(BaseRule):
        """Create subclass of BaseRule without evaluate() method."""

        rule_id = OBJECT_ID
        description = 'This is a rule that properly instantiates BaseRule.'
        severity = 'FAIULRE'
        url = 'http://rule.documentation/'
        resolution = 'This is how you fix this violation.'
        category_tags = {"storage"}
        source_tags = {"stelligent"}

    rule_missing_evaluate = RuleMissingEvaluate(TEMPLATE_MODEL)
    with raises(NotImplementedError,
                match=r"must implement an evaluate\(\) method with logic"):
        rule_missing_evaluate.evaluate()


def test_tags_issue_warning():
    """Sad Path: Show the user a warning if a Rule doesn't have a tag."""
    class TagsMissing(BaseRule):
        """Show warning on subclasses of BaseRule without tags."""

        rule_id = OBJECT_ID
        description = 'This is a rule has no tags.'
        severity = 'FAIULRE'
        url = 'http://rule.documentation/'
        resolution = 'This is how you fix this violation.'

        def evaluate(self):
            pass

    with warns(UserWarning, match=r"tags properties are optional."):
        TagsMissing(TEMPLATE_MODEL)


def test_tag_cleaning():
    """Happy Path: """
    class TagCleaned(BaseRule):
        """Ensure that tags are cleaned and valid."""

        rule_id = OBJECT_ID
        description = 'This is a rule that properly instantiates BaseRule.'
        severity = 'FAIULRE'
        url = 'http://rule.documentation/'
        resolution = 'This Class is valid.'
        category_tags = {"STORAGE"}
        source_tags = {"SteLLigent"}

        def evaluate(self):
            pass

    clean_rule = TagCleaned(TEMPLATE_MODEL)
    assert clean_rule.clean_category_tags == {"storage"}
    assert clean_rule.clean_source_tags == {"stelligent"}

    class TagsWithNumber(BaseRule):
        """Ensure that tags are cleaned and valid."""

        rule_id = OBJECT_ID
        description = 'This is a rule that properly instantiates BaseRule.'
        severity = 'FAIULRE'
        url = 'http://rule.documentation/'
        resolution = 'This Class is valid.'
        category_tags = {"StorAGE", 2600}
        source_tags = {"SteLLigent", 7200}

        def evaluate(self):
            pass

    remove_numbers = TagsWithNumber(TEMPLATE_MODEL)
    assert remove_numbers.clean_category_tags == {"storage"}
    assert remove_numbers.clean_source_tags == {"stelligent"}



"""Tests for BaseParser."""

from pytest import raises
from mc_nag.base_utils.parser import BaseParser
from mc_nag.base_utils.exceptions import InvalidTemplate
from mc_nag.base_utils import read_template_file

TEMPLATES_PATH = 'tests/templates'


def test_base_parser_good_json():
    """Happy Path: Ensure parser can handle a basic template."""
    filepath = f'{TEMPLATES_PATH}/parser_test_good.json'

    with open(filepath) as template:
        template_string = template.read()

    parser = BaseParser(template_string)

    assert len(parser.resources) == 1
    assert parser.resources[0]['name'] == 'OnlyResource'
    with raises(InvalidTemplate):
        parser.validate_file('template', parser.parsed_template,
                            ['MISSING_KEY'])


def test_base_parser_bad_json():
    """Sad Path: Ensure parser throws exception for bad template."""
    filepath = f'{TEMPLATES_PATH}/parser_test_bad.json'

    with open(filepath) as template:
        template_string = template.read()

    with raises(InvalidTemplate, match=r"Invalid YAML or JSON syntax"):
        _ = BaseParser(template_string)


def test_base_parser_bad_line_number():
    """Sad Path: Ensure parser returns -1 when no line number found."""
    filepath = f'{TEMPLATES_PATH}/parser_test_good.json'

    with open(filepath) as template:
        template_string = template.read()

    parser = BaseParser(template_string)

    assert len(parser.resources) == 1
    assert parser.resources[0]['name'] == 'OnlyResource'
    assert parser.find_line_number('DOESNOTEXIST') == -1
    assert not parser.parameters
    assert not parser.outputs
    assert not parser.functions
    assert not parser.variables


def test_parser_bad_syntax_param_file():
    """Sad Path: Ensure parser throws exception for a bad Azure parameter JSON file.."""
    template_string = read_template_file(f'{TEMPLATES_PATH}/' +
                                         'parser_test_good.json')
    parameters_string = read_template_file(f'{TEMPLATES_PATH}/' +
                                           'parser_test_param_file_bad.json')

    with raises(InvalidTemplate, match=r"Invalid YAML or JSON syntax!"):
        _ = BaseParser(template_string, parameters_string)



"""Rule to test the duplicate rules check."""

from mc_nag.base_utils.models.rule import (
    BaseRule,
    WARNING as RULE_WARNING
)


class DuplicateCustomPlatformRule(BaseRule):
    """Create a duplicate custom rule to test duplicate rules check."""

    rule_id = 'custom-platform-rule-test'
    description = """Testing the duplicate rules check."""
    severity = RULE_WARNING
    url = 'https://github.com/stelligent/mc-nag/tests/rules/custom/' + \
          'duplicate_custom_rule_for_testing.py'
    resolution = 'N/A'

    def evaluate(self):
        """Return all resources."""
        return self.template_model.resources



"""Distribution instructions for the package."""

import os
from setuptools import setup, find_packages

requirements_files = ['requirements.txt']
install_requires = []


def read_file(rel_path):
    """Read file contents."""
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, rel_path), 'r') as f_h:
        return f_h.read()


def get_version(rel_path):
    """Retrieve version from passed file."""
    for line in read_file(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


for requirements_file in requirements_files:
    with open(requirements_file, 'r') as f:
        install_requires += f.readlines()

long_description = read_file('README.md')

setup(
    name='mc_nag',
    version=get_version('mc_nag/__init__.py'),
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    install_requires=install_requires,
    url='https://github.com/stelligent/mc-nag',
    license='',
    author='Stelligent',
    author_email='stelligent@stelligent.com',
    description='mc-nag multi-cloud static analysis tool',
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires='>=3.7'
)



"""Printers for outputing results in different formats."""

from json import dumps as json_dumps
from yaml import dump as yaml_dump


BCOLORS = {
    'HEADER': '\033[95m',
    'OKBLUE': '\033[94m',
    'OKGREEN': '\033[92m',
    'WARNING': '\033[93m',
    'ERROR': '\033[91m',
    'ENDC': '\033[0m',
    'BOLD': '\033[1m',
    'UNDERLINE': '\033[4m'
}


class OutputPrinter:
    """Helper class to output violations data in different formats.

    Each method is a static method that accepts the following params:
    :param violations: Dictionary of all violations, keyed by rule object.
    :param counts: Dictionary of counts of violations by severity.
    :param rules: List of rules evaluated against the templates.
    """

    @staticmethod
    def none(*unused):
        """Silence output."""

    @staticmethod
    def json(violations, counts, rules=None):
        """Output aggreagted information in JSON format.

        :param violations: Dict of all violations found keyed by a tuple of
                        (rule_name, rule_id).
        :param counts: Dict of counts of violations keyed by severity.
        :param rules: List of rules evaluated against the templates.
        """
        print(json_dumps(OutputPrinter._build_output(violations, counts, rules)))

    @staticmethod
    def yaml(violations, counts, rules=None):
        """Output aggreagted information in YAML format.

        :param violations: Dict of all violations found keyed by a tuple of
                           (rule_name, rule_id).
        :param counts: Dict of counts of violations keyed by severity.
        :param rules: List of rules evaluated against the templates.
        """
        print(yaml_dump(OutputPrinter._build_output(violations, counts, rules)))

    @staticmethod
    def text(violations, counts, rules=None):
        """Output aggreagted information in plain text format.

        :param violations: Dict of all violations found keyed by a tuple of
                           (rule_name, rule_id).
        :param counts: Dict of counts of violations keyed by severity.
        :param rules: List of rules evaluated against the templates.
        """
        # Print report if any violations
        if counts:
            left_pad = max([len(severity) for severity in counts])
            right_pad = max([len(str(counts[severity]))
                             for severity in counts])

            # Build rule details output
            for rule, violating_resources in violations.items():
                resources_string = [
                    f'{resource.resource_name} (line: {resource.line_number})'
                    for resource in violating_resources]
                header = f'{BCOLORS[rule.severity]}{rule.__name__} ' + \
                    f'({rule.rule_id}){BCOLORS["ENDC"]}'
                left_margin = BCOLORS[rule.severity] + \
                    '|'.rjust(left_pad + 2, ' ') + BCOLORS['ENDC']
                severity = \
                    f"{BCOLORS[rule.severity]}{rule.severity:{left_pad}}" + \
                    BCOLORS['ENDC']
                print(f"""{severity} : {header}
{left_margin}    {rule.description}
{left_margin}    {rule.url}
{left_margin}
{left_margin} Violating resources: {resources_string}
{left_margin}
{left_margin} To resolve:""")
                for line in rule.resolution.split('\n'):
                    print(f'{left_margin} {line}')

            print()
            print(f'{BCOLORS["HEADER"]}{"Report":30}{BCOLORS["ENDC"]}')
            print(''.rjust(30, '-'))
            for severity in sorted(counts):
                print(
                    f'{BCOLORS[severity]}{severity:{left_pad}}{BCOLORS["ENDC"]} : ' +  # noqa: W504
                    f'{counts[severity]:{right_pad}}'
                )
            print(f'\n{BCOLORS["HEADER"]}{"Rules Evaluated":30}{BCOLORS["ENDC"]}')
            print(''.rjust(30, '-'))
            for rule in _sort_rules(rules):
                print(rule.__name__)
        else:
            print(f'{BCOLORS["OKGREEN"]}No violations{BCOLORS["ENDC"]} 🎉')

    @staticmethod
    def _build_output(violations, counts, rules=None):
        """Return data organized into dictionary for output.

        :param violations: Dict of all violations found keyed by a tuple of
                           (rule_name, rule_id).
        :param counts: Dict of counts of violations keyed by severity.
        :param rules: List of rules evaluated against the templates.
        """
        output = {
            'violations': [],
            'report': {
                'rules_evaluated': [rule.__name__ for rule in _sort_rules(rules)],
                'total_rules_evaluated': len(rules),
                'total_violations': 0
            }
        }

        # Build report if any violations
        if counts:
            # Build rule details output
            for rule, violating_resources in violations.items():
                rule_dict = {
                    'name': rule.__name__,
                    'rule_id': rule.rule_id,
                    'severity': rule.severity,
                    'description': rule.description,
                    'url': rule.url,
                    'resolution': rule.resolution,
                    'violating_resources': [
                        resource.__dict__ for resource in violating_resources
                    ]
                }
                output['violations'].append(rule_dict)

            for severity in sorted(counts):
                output['report'][severity] = counts[severity]
                output['report']['total_violations'] += counts[severity]

        return output


def _sort_rules(rules):
    """Return a sorted list of rule names."""
    return sorted(rules, key=lambda x: x.__name__)
